// Synchronize over ssh

#define _GNU_SOURCE

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <poll.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert_ops/libc.h>
#include <bestiola/assert.h>
#include <bestiola/err_wrap.h>

#include "proto.h"

typedef struct {
	FILE *in;
	int out;
	int err;
	char err_buffer[BUFSIZ];
} cmd_io;

typedef struct {
	char *name;
	union {
		const char *link;
		const char *size;
	};
	const char *time;
	const char *mode;
	FILE *stream;
} file_stat;

enum {
	DATA,
	DSH,
	LSH,
	SSH,
	IO_COUNT
};

static const char DDIRS[] = "ddirs";
static const char DFILES[] = "dfiles";
static const char DLINKS[] = "dlinks";
static const char SDIRS[] = "sdirs";
static const char SFILES[] = "sfiles";
static const char SLINKS[] = "slinks";

static char *dtdir;
static const char *findargs = "";
static int first_argn;
static char *from_host;
static cmd_io io[IO_COUNT] = {
	{ .err = -1 }, { .err = -1 }, { .err = -1 }, { .err = -1 }
};
static char *ltdir;
static char *stdir;
static char *to_host;

static void abort_missing_arguments(void);
static void background_cmd(char *argv[], cmd_io *io);
static void call_data_cmd(char *host, char *destdir, cmd_io *io);
static file_stat *create_stat(void);
static void destroy_stat(file_stat *fstat);
static char *escaped_string(const char *s);
static void fatal_error(void);
static void flush_error_streams(void);
static char **get_shell_argv(char *host);
static int is_escape_char(int c);
static void process_args(int argc, char **argv);
static void put_file(const file_stat *fstat);
static void put_int32(int32_t i32);
static void put_link(const file_stat *fstat);
static void put_mode(mode_t mode);
static void put_size(size_t size);
static void put_string(const char *s);
static void read_error_stream(cmd_io *io);
static char *read_line(FILE *stream);
static char *read_line_value(int fd);
static void refill_dir_stat(file_stat *fstat);
static void refill_file_stat(file_stat *fstat);
static void refill_link_stat(file_stat *fstat);
static char *refill_stat_name(file_stat *fstat);
static char *script_init(const cmd_io *io);
static void send_command(const char *command, cmd_io *sh, const char *out);
static void set_cloexec(int fd);
static void set_nonblock(int fd);
static void sync_file(const char *file);
static void update_mode(const file_stat *fstat);
static void update_mode_command(const char *mode, const file_stat *fstat);
static void update_time(const file_stat *fstat);
static void wait_data_available(int fd);

int main(int argc, char **argv) {
	assert_error_handler = fatal_error;

	process_args(argc, argv);

	struct sigaction sa;
	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction(SIGPIPE, &sa, NULL);

	char **lshargv = get_shell_argv(NULL);
	background_cmd(lshargv, &io[LSH]);
	ltdir = script_init(&io[LSH]);

	char **dshargv = get_shell_argv(to_host);
	background_cmd(dshargv, &io[DSH]);
	dtdir = script_init(&io[DSH]);

	char *destdir = argv[argc - 1];
	char *edestdir = escaped_string(destdir);
	assert_fprintf(io[DSH].in, "cd \"%s\"\n", edestdir);

	call_data_cmd(to_host, to_host? edestdir: destdir, &io[DATA]);
	free(edestdir);

	char **sshargv = get_shell_argv(from_host);
	background_cmd(sshargv, &io[SSH]);
	stdir = script_init(&io[SSH]);
	assert_fputs("STARTPWD=\"$PWD\"\n", io[SSH].in);

	int n = argc - 1;
	if (to_host)
		n -= 2;
	int i;
	for (i = first_argn; i < n; i++) {
		char *file_arg = strdup(argv[i]);
		char *last_slash = strrchr(file_arg, '/');
		while (last_slash && !last_slash[1]) {
			*last_slash = '\0';
			last_slash = strrchr(file_arg, '/');
		}
		if (last_slash) {
			*last_slash++ = '\0';
			char *esrcdir = escaped_string(file_arg);
			assert_fprintf(io[SSH].in, "cd \"%s\"\n", esrcdir);
			free(esrcdir);
			sync_file(last_slash);
			assert_fputs("cd \"$STARTPWD\"\n", io[SSH].in);
		}
		else
			sync_file(file_arg);
		free(file_arg);
	}

	free(dtdir);
	free(ltdir);
	free(stdir);

	assert_fputs("echo\n", io[DSH].in);
	assert_fflush(io[DSH].in);
	wait_data_available(io[DSH].out);

	for (i = 0; i < IO_COUNT; i++) {
		assert_fclose(io[i].in);
		assert_close(io[i].out);
	}

	flush_error_streams();

	for (i = 0; i < IO_COUNT; i++)
		assert_close(io[i].err);

	return EXIT_SUCCESS;
}

static void abort_missing_arguments(void) {
	errx_fail("missing arguments");
}

static char *aprint_command(const char *file, char type, const char *format) {
	static const char FIND_FORMAT[] =
		"[ ! -e \"%s\" ] ||"
		" ( { find \"%s\" %s -type %c -printf \"%s\\n\" || [ $? -eq 1 ]"
		" || touch \"$TDIR\"/error ; } | LANG= sort )\n"
		"[ $? -eq 0 -a ! -e \"$TDIR\"/error ] || kill -HUP $$";
	char *efile = escaped_string(file);
	char *strp = assert_aprintf(FIND_FORMAT, efile, efile, findargs, type,
		format);
	free(efile);
	return strp;
}

static void background_cmd(char *argv[], cmd_io *io) {
	int infd[2];
	int outfd[2];
	int errfd[2];
	pipe(infd);
	pipe(outfd);
	pipe(errfd);
	switch (fork()) {
	case 0:
		assert_close(0);
		assert_dup2(infd[0], 0);
		assert_close(1);
		assert_dup2(outfd[1], 1);
		assert_close(2);
		assert_dup2(errfd[1], 2);
		assert_close(infd[0]);
		assert_close(infd[1]);
		assert_close(outfd[0]);
		assert_close(outfd[1]);
		assert_close(errfd[0]);
		assert_close(errfd[1]);
		assert_execvp(argv[0], argv);
	default:
		;
	}
	assert_close(infd[0]);
	assert_close(outfd[1]);
	assert_close(errfd[1]);
	set_cloexec(infd[1]);
	io->in = assert_fdopen(infd[1], "w");
	set_cloexec(outfd[0]);
	set_nonblock(outfd[0]);
	io->out = outfd[0];
	set_cloexec(errfd[0]);
	set_nonblock(errfd[0]);
	io->err = errfd[0];
}

static void call_data_cmd(char *host, char *destdir, cmd_io *io) {
	if (host) {
		char *command = assert_aprintf("sshsync-writer \"%s\"",
			destdir);
		char *argv[] = { "ssh", "-T", "-oClearAllForwardings=yes", host,
			command, NULL };
		background_cmd(argv, io);
		free(command);
		return;
	}
	char *argv[] = { "sshsync-writer", destdir, NULL };
	background_cmd(argv, io);
}

static file_stat *create_stat(void) {
	file_stat *fstat = assert_malloc(sizeof(file_stat));
	fstat->name = NULL;
	fstat->stream = NULL;
	return fstat;
}

static void destroy_stat(file_stat *fstat) {
	free(fstat->name);
	free(fstat);
}

static char *escaped_string(const char *s) {
	int len;
	const char *c;
	for (len = 1, c = s; *c; len++, c++)
		if (is_escape_char(*c))
			len++;
	char *escaped = assert_malloc(len);
	char *e = escaped;
	c = s;
	while (*c) {
		if (is_escape_char(*c))
			*e++ = '\\';
		*e++ = *c++;
	}
	*e = '\0';
	return escaped;
}

static void fatal_error(void) {
	flush_error_streams();
	exit(EXIT_FAILURE);
}

static void flush_error_streams(void) {
	int i;
	for (i = 0; i < IO_COUNT; i++) {
		if (io[i].err != -1)
			read_error_stream(&io[i]);
		if (*io[i].err_buffer)
			assert_fprintf(stderr, "%s\n", io[i].err_buffer);
	}
}

static char **get_shell_argv(char *host) {
	if (host) {
		static char *argv[] = { "ssh", "-T",
			"-oClearAllForwardings=yes", NULL, "sh", NULL };
		argv[3] = host;
		return argv;
	}
	else {
		static char *argv[] = { "sh", NULL };
		return argv;
	}
}

static void get_sync_char(int fd) {
	wait_data_available(fd);
	char c;
	if (read(fd, &c, 1) != 1)
		assert_error_handler();
}

static int is_escape_char(int c) {
	switch (c) {
	case '"':
	case '$':
	case '\\':
	case '`':
		return 1;
	}
	return 0;
}

static FILE *open_tmpfile(const char *name) {
	char *path = assert_aprintf("%s/%s", ltdir, name);
	FILE *stream = assert_fopen(path, "r");
	free(path);
	return stream;
}

static void process_args(int argc, char **argv) {
	switch (argc) {
	case 1:
		abort_missing_arguments();
	case 2:
		if (!strcmp(argv[1], "--findargs")
			|| !strcmp(argv[1], "--from-host")
			|| !strcmp(argv[1], "--to-host"))
			abort_missing_arguments();
		errx(EXIT_FAILURE, "unrecognized option `%s'", argv[1]);
	}

	first_argn = 1;
	if (!strcmp(argv[first_argn], "--findargs")) {
		if (argc == 3)
			abort_missing_arguments();
		findargs = argv[first_argn + 1];
		first_argn += 2;
	}
	if (!strcmp(argv[first_argn], "--from-host")) {
		if (first_argn + 1 >= argc)
			abort_missing_arguments();
		from_host = argv[first_argn + 1];
		first_argn += 2;
	}
	int file_args = argc - first_argn;
	if (!strcmp(argv[argc - 3], "--to-host")) {
		to_host = argv[argc - 2];
		file_args -= 2;
	}
	if (file_args < 2)
		abort_missing_arguments();
}

static void put_file(const file_stat *fstat) {
	assert_putc(PUT, io[DATA].in);
	put_string(fstat->name);

	mode_t mode = assert_strtoul(fstat->mode, 8);
	put_mode(mode);

	size_t size = assert_strtoul(fstat->size, 10);
	put_size(size);
	if (size) {
		char *ename = escaped_string(fstat->name);
		assert_fprintf(io[SSH].in, "cat-expected \"%s\" %s %zu\n",
			ename, fstat->mode, size);
		free(ename);
		assert_fflush(io[SSH].in);
		static char buffer[BUFSIZ];
		while (size) {
			if (BUFSIZ > SSIZE_MAX)
				bad_assertion();
			size_t nread = read(io[SSH].out, buffer, BUFSIZ);
			switch (nread) {
			case (size_t) -1:
				if (errno != EAGAIN)
					fatal_error();
				wait_data_available(io[SSH].out);
				break;
			case 0:
				fatal_error();
				break;
			default:
				assert_fwrite(buffer, 1, nread, io[DATA].in);
				size -= nread;
			}
		}
	}

	struct timespec mtime;
	char *end;
	mtime.tv_sec = strtol(fstat->time, &end, 10);
	if (*end == '.') {
		end++;
		char decimal[10];
		int i;
		for (i = 0; i < 9; i++) {
			if (!end[i])
				break;
			decimal[i] = end[i];
		}
		decimal[i] = '\0';
		long value = strtol(decimal, NULL, 10);
		for (; i < 9; i++)
			value *= 10;
		mtime.tv_nsec = mtime.tv_sec >= 0? value: -value;
	}
	else {
		mtime.tv_nsec = 0;
	}
	put_int32(mtime.tv_sec);
	put_int32(mtime.tv_nsec);
}

static void put_int32(int32_t i32) {
	int i;
	for (i = 0; i < 4; i++) {
		assert_putc(i32 & 0xffu, io[DATA].in);
		i32 >>= 8;
	}
}

static void put_link(const file_stat *fstat) {
	char *ename = escaped_string(fstat->name);
	char *elink = escaped_string(fstat->link);
	assert_fprintf(io[DSH].in, "ln -fns \"%s\" \"%s\"\n", elink, ename);
	free(ename);
	free(elink);
	update_time(fstat);
}

static void put_mode(mode_t mode) {
	int i;
	for (i = 0; i < 2; i++) {
		assert_putc(mode & 0xffu, io[DATA].in);
		mode >>= 8;
	}
}

static void put_size(size_t size) {
	do {
		int c = size & 0x7f;
		size >>= 7;
		if (size)
			c |= 0x80u;
		assert_putc(c, io[DATA].in);
	} while (size);
}

static void put_string(const char *s) {
	assert_fputs(s, io[DATA].in);
	assert_putc('\0', io[DATA].in);
}

static void read_error_stream(cmd_io *io) {
	int err = io->err;
	char *buffer = io->err_buffer;
	for (;;) {
		size_t len = strlen(buffer);
		ssize_t nread = read(err, buffer + len, BUFSIZ - 1 - len);
		switch (nread) {
		case 0:
			return;
		case -1:
			if (errno == EAGAIN)
				return;
			exit(EXIT_FAILURE);
		}
		buffer[len + nread] = '\0';
		char *message = buffer;
		char *c;
		for (c = buffer + len; *c; c++)
			if (*c == '\n') {
				*c = '\0';
				assert_fprintf(stderr, "%s\n", message);
				message = c + 1;
			}
		if (!*message)
			buffer[0] = '\0';
		else if (message == buffer) {
			assert_fputs(buffer, stderr);
			buffer[0] = '\0';
		}
		else
			memmove(buffer, message, strlen(message) + 1);
	}
}

static char *read_line(FILE *stream) {
	size_t size = 128;
	char *line = assert_malloc(size);
	size_t len = 0;
	for (;;) {
		int c = assert_getc(stream);
		if (c == EOF)
			break;
		line[len++] = c;
		if (len == size) {
			size <<= 1;
			line = assert_realloc(line, size);
		}
		if (c == '\n')
			break;
	}
	line[len++] = '\0';
	line = assert_realloc(line, len);
	return line;
}

static char *read_line_value(int fd) {
	wait_data_available(fd);
	FILE *stream = assert_fdopen(dup(fd), "r");
	char *value = read_line(stream);
	assert_fclose(stream);

	size_t len = strlen(value);
	if (len-- < 1)
		return value;
	if (value[len] != '\n')
		return value;
	value[len] = '\0';
	return value;
}

static void refill_dir_stat(file_stat *fstat) {
	char *buffer = refill_stat_name(fstat);
	if (!*buffer)
		return;
	while (*++buffer != '\0');
	fstat->time = ++buffer;
	while (*++buffer != '\0');
	fstat->mode = ++buffer;
	while (*++buffer != '\n');
	*buffer = '\0';
}

static void refill_file_stat(file_stat *fstat) {
	char *buffer = refill_stat_name(fstat);
	if (!*buffer)
		return;
	while (*++buffer != '\0');
	fstat->size = ++buffer;
	while (*++buffer != '\0');
	fstat->time = ++buffer;
	while (*++buffer != '\0');
	fstat->mode = ++buffer;
	while (*++buffer != '\n');
	*buffer = '\0';
}

static void refill_link_stat(file_stat *fstat) {
	char *buffer = refill_stat_name(fstat);
	if (!*buffer)
		return;
	while (*++buffer != '\0');
	fstat->link = ++buffer;
	while (*++buffer != '\0');
	fstat->time = ++buffer;
	while (*++buffer != '\n');
	*buffer = '\0';
}

static char *refill_stat_name(file_stat *fstat) {
	free(fstat->name);
	char *buffer = read_line(fstat->stream);
	fstat->name = buffer;
	return buffer;
}

static char *script_init(const cmd_io *io) {
	assert_fputs(
		"set -e\n"
		"TDIR=\n"
		"cleanup() {\n"
			"rm -fr \"$TDIR\"\n"
		"}\n"
		"trap cleanup EXIT\n"
		"TDIR=$(mktemp -d)\n"
		"echo \"$TDIR\"\n", io->in);
	assert_fflush(io->in);

	char *tdir = read_line_value(io->out);
	if (!strlen(tdir))
		assert_error_handler();
	return tdir;
}

static void send_command(const char *command, cmd_io *sh, const char *out) {
	assert_fprintf(sh->in,
		"{ ( %s\n ) || touch \"$TDIR\"/error ; } | gzip -9n"
		" > \"$TDIR\"/out.gz\n"
		"[ ! -e \"$TDIR\"/error ]\n"
		"stat -c %%s \"$TDIR\"/out.gz\n", command);
	assert_fflush(sh->in);

	char *tsize_str = read_line_value(sh->out);
	size_t size = assert_strtoul(tsize_str, 10);
	free(tsize_str);

	char *path = assert_aprintf("%s/out.gz", ltdir);
	FILE *stream = assert_fopen(path, "w");
	free(path);

	assert_fputs("cat \"$TDIR\"/out.gz\n", sh->in);
	assert_fflush(sh->in);
	static char buffer[BUFSIZ];
	while (size) {
		if (BUFSIZ > SSIZE_MAX)
			bad_assertion();
		size_t nread = read(sh->out, buffer, BUFSIZ);
		switch (nread) {
		case (size_t) -1:
			if (errno != EAGAIN)
				fatal_error();
			wait_data_available(sh->out);
			break;
		case 0:
			fatal_error();
			break;
		default:
			assert_fwrite(buffer, 1, nread, stream);
			size -= nread;
		}
	}

	assert_fclose(stream);
	assert_fprintf(io[LSH].in,
		"gunzip -c \"$TDIR\"/out.gz > \"$TDIR\"/%s\n", out);

	assert_fputs("echo\n", io[LSH].in);
	assert_fflush(io[LSH].in);
	get_sync_char(io[LSH].out);
}

static void set_cloexec(int fd) {
	int flags = assert_fcntl(fd, F_GETFD);
	assert_fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
}

static void set_nonblock(int fd) {
	int flags = assert_fcntl(fd, F_GETFL);
	assert_fcntl(fd, F_SETFL, flags | O_NONBLOCK);
}

static void sync_file(const char *file) {
	char *last_deleted = NULL;
	file_stat *dfile = create_stat();
	file_stat *sfile = create_stat();

	char *list_cmd = aprint_command(file, 'd', "%p\\0%T@\\0%#m");
	send_command(list_cmd, &io[DSH], DDIRS);
	send_command(list_cmd, &io[SSH], SDIRS);
	free(list_cmd);

	FILE *ddirs = open_tmpfile(DDIRS);
	FILE *sdirs = open_tmpfile(SDIRS);

	dfile->stream = ddirs;
	refill_dir_stat(dfile);
	sfile->stream = sdirs;
	refill_dir_stat(sfile);
	while (*dfile->name) {
		if (!*sfile->name || strcmp(dfile->name, sfile->name) < 0) {
			if (!last_deleted || strncmp(dfile->name, last_deleted,
				strlen(last_deleted))) {
				printf("deleting %s/\n", dfile->name);
				char *ename = escaped_string(dfile->name);
				assert_fprintf(io[DSH].in, "rm -r \"%s\"\n",
					ename);
				free(ename);
				free(last_deleted);
				last_deleted = assert_aprintf("%s/",
					dfile->name);
			}
			refill_dir_stat(dfile);
		}
		else if (!strcmp(dfile->name, sfile->name)) {
			refill_dir_stat(sfile);
			refill_dir_stat(dfile);
		}
		else
			refill_dir_stat(sfile);
	}
	free(last_deleted);

	list_cmd = aprint_command(file, 'f', "%p\\0%s\\0%T@\\0%#m");
	send_command(list_cmd, &io[DSH], DFILES);
	send_command(list_cmd, &io[SSH], SFILES);
	free(list_cmd);

	FILE *dfiles = open_tmpfile(DFILES);
	FILE *sfiles = open_tmpfile(SFILES);

	dfile->stream = dfiles;
	refill_file_stat(dfile);
	sfile->stream = sfiles;
	refill_file_stat(sfile);
	while (*dfile->name) {
		if (!*sfile->name || strcmp(dfile->name, sfile->name) < 0) {
			printf("deleting %s\n", dfile->name);
			char *ename = escaped_string(dfile->name);
			assert_fprintf(io[DSH].in, "rm \"%s\"\n", ename);
			free(ename);
			refill_file_stat(dfile);
		}
		else if (!strcmp(dfile->name, sfile->name)) {
			refill_file_stat(sfile);
			refill_file_stat(dfile);
		}
		else
			refill_file_stat(sfile);
	}

	list_cmd = aprint_command(file, 'l', "%p\\0%l\\0%T@");
	send_command(list_cmd, &io[DSH], DLINKS);
	send_command(list_cmd, &io[SSH], SLINKS);
	free(list_cmd);

	FILE *dlinks = open_tmpfile(DLINKS);
	FILE *slinks = open_tmpfile(SLINKS);

	dfile->stream = dlinks;
	refill_link_stat(dfile);
	sfile->stream = slinks;
	refill_link_stat(sfile);
	while (*dfile->name) {
		if (!*sfile->name || strcmp(dfile->name, sfile->name) < 0) {
			printf("deleting %s -> %s\n", dfile->name, dfile->link);
			char *ename = escaped_string(dfile->name);
			assert_fprintf(io[DSH].in, "rm \"%s\"\n", ename);
			free(ename);
			refill_link_stat(dfile);
		}
		else if (!strcmp(dfile->name, sfile->name)) {
			refill_link_stat(sfile);
			refill_link_stat(dfile);
		}
		else
			refill_link_stat(sfile);
	}

	rewind(ddirs);
	rewind(sdirs);

	dfile->stream = ddirs;
	refill_dir_stat(dfile);
	sfile->stream = sdirs;
	refill_dir_stat(sfile);
	while (*sfile->name) {
		if (!*dfile->name || strcmp(dfile->name, sfile->name) > 0) {
			printf("adding %s/\n", sfile->name);
			char *ename = escaped_string(sfile->name);
			assert_fprintf(io[DSH].in, "mkdir \"%s\"\n", ename);
			free(ename);
			update_mode(sfile);
			refill_dir_stat(sfile);
		}
		else if (!strcmp(dfile->name, sfile->name)) {
			if (strcmp(dfile->mode, sfile->mode)) {
				printf("updating %s/\n", sfile->name);
				update_mode(sfile);
			}
			refill_dir_stat(sfile);
			refill_dir_stat(dfile);
		}
		else
			refill_dir_stat(dfile);
	}

	assert_fputs("echo\n", io[DSH].in);
	assert_fflush(io[DSH].in);
	get_sync_char(io[DSH].out);

	rewind(dfiles);
	rewind(sfiles);

	dfile->stream = dfiles;
	refill_file_stat(dfile);
	sfile->stream = sfiles;
	refill_file_stat(sfile);
	bool file_put = false;
	while (*sfile->name) {
		if (!*dfile->name || strcmp(dfile->name, sfile->name) > 0) {
			printf("adding %s\n", sfile->name);
			put_file(sfile);
			file_put = true;
			refill_file_stat(sfile);
		}
		else if (!strcmp(dfile->name, sfile->name)) {
			if (strcmp(dfile->size, sfile->size)
				|| strcmp(dfile->time, sfile->time)) {
				printf("updating %s\n", sfile->name);
				put_file(sfile);
				file_put = true;
			}
			else if (strcmp(dfile->mode, sfile->mode)) {
				printf("updating %s\n", sfile->name);
				update_mode(sfile);
			}
			refill_file_stat(sfile);
			refill_file_stat(dfile);
		}
		else
			refill_file_stat(dfile);
	}
	if (file_put) {
		assert_putc(SYNC, io[DATA].in);
		assert_fflush(io[DATA].in);
	}

	assert_fclose(dfiles);
	assert_fclose(sfiles);

	rewind(dlinks);
	rewind(slinks);

	dfile->stream = dlinks;
	refill_link_stat(dfile);
	sfile->stream = slinks;
	refill_link_stat(sfile);
	while (*sfile->name) {
		if (!*dfile->name || strcmp(dfile->name, sfile->name) > 0) {
			printf("adding %s -> %s\n", sfile->name, sfile->link);
			put_link(sfile);
			refill_link_stat(sfile);
		}
		else if (!strcmp(dfile->name, sfile->name)) {
			if (strcmp(dfile->link, sfile->link)
				|| strcmp(dfile->time, sfile->time)) {
				printf("updating %s -> %s\n", sfile->name,
					sfile->link);
				put_link(sfile);
			}
			refill_link_stat(sfile);
			refill_link_stat(dfile);
		}
		else
			refill_link_stat(dfile);
	}

	assert_fclose(dlinks);
	assert_fclose(slinks);

	if (file_put) {
		get_sync_char(io[DATA].out);

		assert_fclose(ddirs);
		list_cmd = aprint_command(file, 'd', "%p\\0%T@\\0%#m");
		send_command(list_cmd, &io[DSH], DDIRS);
		free(list_cmd);
		ddirs = open_tmpfile(DDIRS);
	}
	else
		rewind(ddirs);

	rewind(sdirs);

	dfile->stream = ddirs;
	refill_dir_stat(dfile);
	sfile->stream = sdirs;
	refill_dir_stat(sfile);
	while (*sfile->name) {
		if (!*dfile->name || strcmp(dfile->name, sfile->name) > 0) {
			printf("updating %s/\n", sfile->name);
			update_time(sfile);
			refill_dir_stat(sfile);
		}
		else if (!strcmp(dfile->name, sfile->name)) {
			if (strcmp(dfile->time, sfile->time)) {
				printf("updating %s/\n", sfile->name);
				update_time(sfile);
			}
			refill_dir_stat(sfile);
			refill_dir_stat(dfile);
		}
		else
			refill_dir_stat(dfile);
	}

	assert_fclose(ddirs);
	assert_fclose(sdirs);

	destroy_stat(dfile);
	destroy_stat(sfile);
}

static void update_mode_command(const char *mode, const file_stat *fstat) {
	char *ename = escaped_string(fstat->name);
	assert_fprintf(io[DSH].in, "chmod g-s \"%s\"\n" "chmod %s \"%s\"\n",
		ename, mode, ename);
	free(ename);
}

static void update_mode(const file_stat *fstat) {
	update_mode_command(fstat->mode, fstat);
}

static void update_time(const file_stat *fstat) {
	char *ename = escaped_string(fstat->name);
	assert_fprintf(io[DSH].in, "touch -hm -d @%s \"%s\"\n", fstat->time,
		ename);
	free(ename);
}

static void wait_data_available(int fd) {
	struct pollfd pollin[IO_COUNT + 1];
	int i;
	for (i = 0; i < IO_COUNT; i++) {
		pollin[i].fd = io[i].err;
		pollin[i].events = POLLIN;
	}
	pollin[IO_COUNT].fd = fd;
	pollin[IO_COUNT].events = POLLIN;

	for (;;) {
		assert_poll(pollin, IO_COUNT + 1, -1);
		int i;
		for (i = 0; i < IO_COUNT; i++) {
			if (pollin[i].revents & POLLIN)
				read_error_stream(&io[i]);
		}
		if (pollin[IO_COUNT].revents & POLLIN)
			break;
		if (pollin[IO_COUNT].revents & POLLHUP)
			fatal_error();
	}
}
