// Perform write operations for sshsync

#define _GNU_SOURCE

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert_ops/libc.h>
#include <bestiola/assert.h>
#include <sys/stat.h>

#include "proto.h"

#define MIN(x, y) ({ \
	typeof(x) __x = (x); \
	typeof(y) __y = (y); \
	__x < __y ? __x : __y; \
})

static char *get_string(void);
static void process_args(int argc);
static void put_file(void);

int main(int argc, char **argv) {
	process_args(argc);
	assert_chdir(argv[1]);

	for (;;) {
		int command = assert_getchar();
		if (command == EOF)
			break;
		switch (command) {
		case PUT:
			put_file();
			break;
		case SYNC:
			assert_putchar('\0');
			assert_fflush(stdout);
			break;
		default:
			errx(EXIT_FAILURE, "Unsupported command");
		}
	}

	return EXIT_SUCCESS;
}

static int32_t get_int32(void) {
	int32_t i = 0;
	int shift;
	for (shift = 0; shift < 32; shift += 8)
		i |= assert_getchar_noeof() << shift;
	return i;
}

static mode_t get_mode(void) {
	mode_t mode = assert_getchar_noeof();
	mode |= assert_getchar_noeof() << 8;
	return mode;
}

static size_t get_size(void) {
	size_t size = 0;
	int shift = 0;
	int c;
	do {
		c = assert_getchar_noeof();
		bool too_big = false;
		switch (sizeof(size_t)) {
		case 4:
			too_big = shift == 28 && c & 0xf0;
			break;
		case 8:
			too_big = shift == 63 && c & 0xfe;
			break;
		default:
			bad_assertion();
		}
		if (too_big)
			errx(EXIT_FAILURE, "Size too big");
		size |= (size_t)(c & 0x7f) << shift;
		shift += 7;
	} while (c & 0x80);
	return size;
}

static char *get_string(void) {
	size_t size = 128;
	char *s = assert_malloc(size);
	size_t len = 0;
	for (;;) {
		int c = assert_getchar_noeof();
		s[len++] = c;
		if (!c) {
			s = assert_realloc(s, len);
			return s;
		}
		if (len == size) {
			size <<= 1;
			s = assert_realloc(s, size);
		}
	}
}

static void process_args(int argc) {
	if (argc == 2)
		return;
	errx(EXIT_FAILURE, "This program is an internal tool.");
}

static void put_file(void) {
	char *path = get_string();
	mode_t mode = get_mode();
	errno = 0;
	if (unlink(path) && errno != ENOENT)
		err(EXIT_FAILURE, NULL);
	int fd = assert_open(path, O_WRONLY | O_CREAT | O_EXCL | O_NOFOLLOW,
		mode | S_IWUSR);
	free(path);
	FILE *stream = assert_fdopen(fd, "w");

	static char buffer[BUFSIZ];
	size_t size = get_size();
	while (size) {
		size_t nmemb = MIN(size, (size_t) BUFSIZ);
		size_t nread = assert_fread(buffer, 1, nmemb, stdin);
		assert_fwrite(buffer, 1, nread, stream);
		if (nread < nmemb)
			err(EXIT_FAILURE, "Unexpected end of file");
		size -= nread;
	}
	assert_fflush(stream);

	if (!(mode & S_IWUSR))
		assert_fchmod(fd, mode);

	struct timespec times[2];
	times[0].tv_nsec = UTIME_OMIT;
	times[1].tv_sec = get_int32();
	times[1].tv_nsec = get_int32();
	assert_futimens(fd, times);

	assert_fclose(stream);
}
